@@ -5,6 +5,7 @@ import six
 
 from . import base_manager
 from . import packet
+from . import namespace
 
 
 class Server(object):
@@ -29,6 +30,9 @@ class Server(object):
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
                  versions.
+    :param async_handlers: If set to ``True``, event handlers are executed in
+                           separate threads. To run handlers synchronously,
+                           set to ``False``. The default is ``False``.
     :param kwargs: Connection parameters for the underlying Engine.IO server.
 
     The Engine.IO configuration supports the following settings:
@@ -61,7 +65,7 @@ class Server(object):
                             ``False``.
     """
     def __init__(self, client_manager=None, logger=False, binary=False,
-                 json=None, **kwargs):
+                 json=None, async_handlers=False, **kwargs):
         engineio_options = kwargs
         engineio_logger = engineio_options.pop('engineio_logger', None)
         if engineio_logger is not None:
@@ -69,6 +73,7 @@ class Server(object):
         if json is not None:
             packet.Packet.json = json
             engineio_options['json'] = json
+        engineio_options['async_handlers'] = False
         self.eio = engineio.Server(**engineio_options)
         self.eio.on('connect', self._handle_eio_connect)
         self.eio.on('message', self._handle_eio_message)
@@ -77,10 +82,9 @@ class Server(object):
 
         self.environ = {}
         self.handlers = {}
+        self.namespace_handlers = {}
 
-        self._binary_packet = None
-        self._attachment_count = 0
-        self._attachments = []
+        self._binary_packet = []
 
         if not isinstance(logger, bool):
             self.logger = logger
@@ -99,6 +103,8 @@ class Server(object):
         self.manager = client_manager
         self.manager_initialized = False
 
+        self.async_handlers = async_handlers
+
         self.async_mode = self.eio.async_mode
 
     def on(self, event, handler=None, namespace=None):
@@ -138,6 +144,8 @@ class Server(object):
         client's acknowledgement callback function if it exists. The
         ``'disconnect'`` handler does not take a second argument.
         """
+        if '-' in event:
+            raise ValueError('event names cannot contain hypens')
         namespace = namespace or '/'
 
         def set_handler(handler):
@@ -150,6 +158,19 @@ class Server(object):
             return set_handler
         set_handler(handler)
 
+    def register_namespace(self, namespace_handler):
+        """Register a namespace handler object.
+
+        :param namespace_handler: An instance of a :class:`Namespace`
+                                  subclass that handles all the event traffic
+                                  for a namespace.
+        """
+        if not isinstance(namespace_handler, namespace.Namespace):
+            raise ValueError('Not a namespace instance')
+        namespace_handler._set_server(self)
+        self.namespace_handlers[namespace_handler.namespace] = \
+            namespace_handler
+
     def emit(self, event, data=None, room=None, skip_sid=None, namespace=None,
              callback=None):
         """Emit a custom event to one or more connected clients.
@@ -384,8 +405,8 @@ class Server(object):
             namespace_list = [namespace]
         for n in namespace_list:
             if n != '/' and self.manager.is_connected(sid, n):
-                self.manager.disconnect(sid, n)
                 self._trigger_event('disconnect', n, sid)
+                self.manager.disconnect(sid, n)
         if namespace == '/' and self.manager.is_connected(sid, namespace):
             self._trigger_event('disconnect', '/', sid)
             self.manager.disconnect(sid, '/')
@@ -397,11 +418,20 @@ class Server(object):
         namespace = namespace or '/'
         self.logger.info('received event "%s" from %s [%s]', data[0], sid,
                          namespace)
-        r = self._trigger_event(data[0], namespace, sid, *data[1:])
+        if self.async_handlers:
+            self.start_background_task(self._handle_event_internal, self, sid,
+                                       data, namespace, id)
+        else:
+            self._handle_event_internal(self, sid, data, namespace, id)
+
+    def _handle_event_internal(self, server, sid, data, namespace, id):
+        r = server._trigger_event(data[0], namespace, sid, *data[1:])
         if id is not None:
             # send ACK packet with the response returned by the handler
             # tuples are expanded as multiple arguments
-            if isinstance(r, tuple):
+            if r is None:
+                data = []
+            elif isinstance(r, tuple):
                 data = list(r)
             else:
                 data = [r]
@@ -409,10 +439,10 @@ class Server(object):
                 binary = False  # pragma: nocover
             else:
                 binary = None
-            self._send_packet(sid, packet.Packet(packet.ACK,
-                                                 namespace=namespace,
-                                                 id=id, data=data,
-                                                 binary=binary))
+            server._send_packet(sid, packet.Packet(packet.ACK,
+                                                   namespace=namespace,
+                                                   id=id, data=data,
+                                                   binary=binary))
 
     def _handle_ack(self, sid, namespace, id, data):
         """Handle ACK packets from the client."""
@@ -422,9 +452,15 @@ class Server(object):
 
     def _trigger_event(self, event, namespace, *args):
         """Invoke an application event handler."""
+        # first see if we have an explicit handler for the event
         if namespace in self.handlers and event in self.handlers[namespace]:
             return self.handlers[namespace][event](*args)
 
+        # or else, forward the event to a namepsace handler if one exists
+        elif namespace in self.namespace_handlers:
+            return self.namespace_handlers[namespace].trigger_event(
+                event, *args)
+
     def _handle_eio_connect(self, sid, environ):
         """Handle the Engine.IO connection event."""
         self.environ[sid] = environ
@@ -432,22 +468,14 @@ class Server(object):
 
     def _handle_eio_message(self, sid, data):
         """Dispatch Engine.IO messages."""
-        if self._attachment_count > 0:
-            self._attachments.append(data)
-            self._attachment_count -= 1
-
-            if self._attachment_count == 0:
-                self._binary_packet.reconstruct_binary(self._attachments)
-                if self._binary_packet.packet_type == packet.BINARY_EVENT:
-                    self._handle_event(sid, self._binary_packet.namespace,
-                                       self._binary_packet.id,
-                                       self._binary_packet.data)
+        if len(self._binary_packet):
+            pkt = self._binary_packet[0]
+            if pkt.add_attachment(data):
+                self._binary_packet.pop(0)
+                if pkt.packet_type == packet.BINARY_EVENT:
+                    self._handle_event(sid, pkt.namespace, pkt.id, pkt.data)
                 else:
-                    self._handle_ack(sid, self._binary_packet.namespace,
-                                     self._binary_packet.id,
-                                     self._binary_packet.data)
-                self._binary_packet = None
-                self._attachments = []
+                    self._handle_ack(sid, pkt.namespace, pkt.id, pkt.data)
         else:
             pkt = packet.Packet(encoded_packet=data)
             if pkt.packet_type == packet.CONNECT:
@@ -460,9 +488,7 @@ class Server(object):
                 self._handle_ack(sid, pkt.namespace, pkt.id, pkt.data)
             elif pkt.packet_type == packet.BINARY_EVENT or \
                     pkt.packet_type == packet.BINARY_ACK:
-                self._binary_packet = pkt
-                self._attachments = []
-                self._attachment_count = pkt.attachment_count
+                self._binary_packet.append(pkt)
             elif pkt.packet_type == packet.ERROR:
                 raise ValueError('Unexpected ERROR packet.')
             else:
